#ifndef _H_LIBCX_FORMAT_FMT
#define _H_LIBCX_FORMAT_FMT

#include <cx/collections/vec.hpp>
#include <stddef.h>
#include <cx/stl.hpp>
#include <string.h>
#include <cx/string.hpp>
#include <cx/format/base_types.hpp>

template <typename T>
concept Format = requires(T t) {
    { t.fmt_to_str() } -> stl::same_as<String>;
} || requires(T t) {
    { Formatter<T>::fmt_to_str(t) } -> stl::same_as<String>;
};

#ifndef __CONCEPT_ONLY

namespace {
    String format(Vec<String>& args, String& _fmt) {
        const char* fmt = _fmt.as_cstr();
        String str = String();
        size_t args_idx = 0;
        for (size_t i = 0; i < strlen(fmt); i++) {
            if (fmt[i] == '{') {
                if (fmt[i + 1] == '{') {
                    str.push('{');
                    str.push('{');
                    i += 1;
                    continue;
                }
                if (fmt[i + 1] == '}') {
                    i += 1;
                    if (args.len() <= args_idx) {
                    }
                    String* arg = args.nth(args_idx).expect("Amount of arguments dont match amount of placeholders in format string");
                    str.push_str(*arg);
                    args_idx += 1;
                }
                continue;
            }
            str.push(fmt[i]);
        }
        return str;
    }

    template <Format T, typename... Types>
    String format(Vec<String>& args, String& fmt, T var1, Types... var2) {
        using ActualType = stl::remove_cvref_t<decltype(var1)>;

        if constexpr (requires(ActualType v) { v.fmt_to_str(); }) {
            args.push(var1.fmt_to_str());
        } else {
            args.push(Formatter<ActualType>::fmt_to_str(var1));
        }

        return format(args, fmt, var2...);
    }
} // namespace

template <typename... Types>
String format(const char* fmt, Types... vars) {
    Vec<String> args = Vec<String>();
    String fmt_ = String(fmt);
    return format(args, fmt_, vars...);
}

// Has to be here, cause c++
template <Format T>
struct Formatter<Vec<T>> {
    static String fmt_to_str(Vec<T>& val) {
        String s = String();
        s.push('[');
        if (val.len() == 0) {
            s.push(']');
            return s;
        }
        String tmp1 = format("{}", *val.nth(0).unwrap()).fmt_to_str();
        s.push_str(tmp1);
        for (size_t i = 1; i < val.len(); i++) {
            s.push_str(", ");

            String tmp2 = format("{}", *val.nth(i).unwrap());
            s.push_str(tmp2);
        }
        s.push(']');
        return s;
    }
};

#endif // __CONCEPT_ONLY
#endif // _H_LIBCX_FORMAT_FMT