
#ifndef _H_LIBCX_OPTION
#define _H_LIBCX_OPTION
#include <cx/stl.hpp>

extern "C" {
int puts(const char* s);
}

template <typename T>
class Option {
    bool has_value = false;
    T storage;

public:
    Option() = default;

    Option(const T& val) {
        this->has_value = true;
        this->storage = stl::move(val);
    }

    Option(T&& val) {
        this->has_value = true;
        this->storage = stl::move(val);
    }

    ~Option() {
        if (has_value) {
            storage.~T();
        }
    }

    bool is_some() const {
        return has_value;
    }

    bool is_none() const {
        return !has_value;
    }

    T& expect(const char* str) {
        if (!has_value) {
            puts("PANIC: Unwrapped a None Option: ");
            puts(str);
            puts("\n");
            __builtin_trap();
        }
        return this->storage;
    }

    T& unwrap() {
        if (!has_value) {
            puts("PANIC: Unwrapped a None Option\n");
            __builtin_trap();
        }
        return this->storage;
    }

    const T& unwrap() const {
        if (!has_value) {
            __builtin_trap();
        }
        return &this->storage;
    }

    T* operator->() {
        return &this->storage;
    }

    T& operator*() {
        return &this->storage;
    }
};

#endif // _H_LIBCX_OPTION