#ifndef _H_LIBCX_STL_IS_INTEGRAL
#define _H_LIBCX_STL_IS_INTEGRAL

#include <cx/stl/integral_constant.hpp>
#include <cx/stl/remove_cvref.hpp>

// clang-format off
template <class _Tp> struct __libcpp_is_integral                     { enum { value = 0 }; };
template <>          struct __libcpp_is_integral<bool>               { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<char>               { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<signed char>        { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<unsigned char>      { enum { value = 1 }; };
#if _LIBCPP_HAS_WIDE_CHARACTERS
template <>          struct __libcpp_is_integral<wchar_t>            { enum { value = 1 }; };
#endif
#if _LIBCPP_HAS_CHAR8_T
template <>          struct __libcpp_is_integral<char8_t>            { enum { value = 1 }; };
#endif
template <>          struct __libcpp_is_integral<char16_t>           { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<char32_t>           { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<short>              { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<unsigned short>     { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<int>                { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<unsigned int>       { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<long>               { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<unsigned long>      { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<long long>          { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<unsigned long long> { enum { value = 1 }; };
#if _LIBCPP_HAS_INT128
template <>          struct __libcpp_is_integral<__int128_t>         { enum { value = 1 }; };
template <>          struct __libcpp_is_integral<__uint128_t>        { enum { value = 1 }; };
#endif
// clang-format on

#if __has_builtin(__is_integral)

template <class _Tp>
struct is_integral : _BoolConstant<__is_integral(_Tp)> {};

template <class _Tp>
inline constexpr bool is_integral_v = __is_integral(_Tp);

#else

template <class _Tp>
struct is_integral : public _BoolConstant<__libcpp_is_integral<__remove_cvref_t<_Tp>>::value> {};

template <class _Tp>
inline constexpr bool is_integral_v = is_integral<_Tp>::value;

#endif // __has_builtin(__is_integral)

#endif // !_H_LIBCX_STL_IS_INTEGRAL
