#ifndef _H_LIBCX_SYNC_SPINLOCK
#define _H_LIBCX_SYNC_SPINLOCK
#include <cx/io/print.hpp>
#include <stdatomic.h>

template <typename T>
struct SpinLock {
private:
    volatile atomic_bool locked = ATOMIC_BOOL_LOCK_FREE;
    T value;

    SpinLock() = default;

    void acquire() {
        while (atomic_load_explicit(&locked, memory_order_acquire) == true) {
            // println("Locked");
        }
    }

    void release() {
        atomic_store_explicit(&locked, false, memory_order_release);
    }

    struct SpinGuard {
    private:
        SpinLock<T>* parent;

    public:
        SpinGuard(SpinLock<T>* p) : parent(p) {
            parent->acquire();
        }

        ~SpinGuard() {
            parent->release();
        }

        T& get() {
            return parent->value;
        }

        T* operator->() {
            return &parent->value;
        }

        T& operator*() {
            return parent->value;
        }

        SpinGuard(const SpinGuard&) = delete;
        SpinGuard& operator=(const SpinGuard&) = delete;
    };

public:
    template <typename... Args>
    explicit SpinLock(Args&&... args) : value((Args&&)args...) {
        atomic_store_explicit(&locked, false, memory_order_release);
    }

    SpinGuard lock() {
        return SpinGuard(this);
    }
};

#endif // !_H_LIBCX_SYNC_SPINLOCK
